#include <htc.h>
#include <stdio.h>
#include "limits.h"

// Chip Konfiguration
__CONFIG(
	FCMDIS &  // Fail Clock Monitor Enable 
	IESODIS &  // Internal External Switch Over 
	BORDIS &  // Brown-out detect modes 
	UNPROTECT &  // Protection of data block 
	UNPROTECT &  // Protection of program code 
	MCLRDIS &  // Master clear reset
	PWRTEN &  // Power up timer enable 
	WDTDIS &  // Watchdog timer enable 
	INTIO  // Oscillator configurations: internal, IO-Funktion statt Clkin/Clkout
);

// ist normalerweise in stdint.h enthalten, was der Hi-Tech Compiler aber nicht kennt
typedef signed   char         int8_t;
typedef unsigned char        uint8_t;
typedef signed   int         int16_t;
typedef unsigned int        uint16_t;
typedef signed   short long  int24_t;	// Microchip specific
typedef unsigned short long uint24_t;	// Microchip specific
typedef signed   long        int32_t;
typedef unsigned long       uint32_t;

// PWM-Werte
static const uint16_t g_pwm[] =
{
	20, 21, 21, 22, 23, 24, 24, 25,
	26, 27, 28, 29, 30, 31, 32, 33,
	34, 35, 36, 37, 38, 40, 41, 42,
	44, 45, 47, 48, 50, 51, 53, 55,
	57, 58, 60, 62, 64, 67, 69, 71,
	73, 76, 78, 81, 84, 86, 89, 92,
	95, 98, 102, 105, 108, 112, 116, 120,
	124, 128, 132, 136, 141, 145, 150, 155,
	160, 166, 171, 177, 183, 189, 195, 201,
	208, 215, 222, 229, 237, 245, 253, 261,
	270, 279, 288, 297, 307, 317, 328, 339,
	350, 361, 373, 386, 398, 412, 425, 439,
	454, 469, 484, 500, 517, 534, 551, 570,
	589, 608, 628, 649, 670, 692, 715, 739,
	763, 789, 815, 842, 869, 898, 928, 959,
	990, 1023
};

// main Funktion
void main(void)
{
	// Ports initialisieren: alles auf Ausgang
	GPIO = 0;
	TRISIO = 0;

	// Oszillator auf 8 MHz einstellen
	OSCCON = (1 << 6) | (1 << 5) | (1 << 4) | 1;

	// Timer0 Counter zurcksetzen
	TMR0 = 0;

	// keine Pullups, Fosc/4 als Taktquelle fr Timer0, 1:32 Prescaler:
	// ergibt bei 8 MHz also Fosc/4/32/256 = 244 Hz
	OPTION = (1 << 7) | (1 << 2);

	// Period Register von Timer2 auf Maximum
	PR2 = 255;

	// Timer2 on, Prescaler 1
	// zusammen mit PR2 ergibt das eine Frequenz von 8 MHz/4/(PR2+1)=4 MHz / 256 = 7812.5 Hz
	T2CON = 1 << 2;

	// keine Interrupts fr Timer2
	PIE1 = 0;

	// PWM-Mode fr das CCP-Modul aktivieren
	CCP1CON = 15;

	// global interrupt enable, Timer0 overflow enable und Interrupt Flag lschen
	INTCON = (1 << 7) | (1 << 5);

	// der Rest luft im Interrupt ab
	while (1) {
	}
}

// Interruptroutine
void interrupt timerInterrupt(void)
{
	// LED mit ca. 1 Hz blinken lassen
	static uint8_t counter = 0;
	static uint8_t value = 0;
	if (counter++ == 122) {
		GPIO = value;
		value = 255 - value;
		counter = 0;
	}

	// Tastverhltnis vom PWM-Ausgang aktualisieren
	static uint8_t pwmIndex = 0;
	static uint8_t direction = 0;
	uint16_t pwm = g_pwm[pwmIndex];
	CCPR1L = (pwm >> 2) & 0xff;
	CCP1CON = 15 | ((pwm & 3) << 4);

	// PWM-Index aktualisieren: zhlt einmal rauf und dann wieder runter
	if (direction) {
		if (pwmIndex == sizeof(g_pwm) / sizeof(uint16_t) - 1) {
			direction = 0;
		} else {
			pwmIndex++;
		}
	} else {
		if (pwmIndex == 0) {
			direction = 1;
		} else {
			pwmIndex--;
		}
	}

	// Interruptflag zurcksetzen
	T0IF = 0;
}
